package cd.casic.ci.common.pipeline.utils;

import cd.casic.framework.commons.exception.ServerException;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.node.ObjectNode;
import com.networknt.schema.JsonSchema;
import com.networknt.schema.JsonSchemaFactory;
import com.networknt.schema.SpecVersion;
import com.networknt.schema.ValidationMessage;
import org.yaml.snakeyaml.Yaml;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

public class MatrixContextUtils {

    private static final String strategyJsonPattern = "^(\\$\\{\\{[ ]*fromJSON\\()([^()]+)(\\)[ ]*}})$";

    private static final String schemaJson = "{\n" +
            "    \"type\": \"object\",\n" +
            "    \"required\": [],\n" +
            "    \"properties\": {\n" +
            "        \"include\": {\n" +
            "            \"description\": \"值格式为：List,用于给 matrix 的指定组合增加额外的属性,或者新增1个或多个组合,每个元素为一个 Object,或是一个'$'{{fromJSON(xxx)}}上下文\",\n" +
            "            \"oneOf\": [\n" +
            "                {\n" +
            "                    \"type\": \"array\",\n" +
            "                    \"items\": {\n" +
            "                        \"type\": \"object\",\n" +
            "                        \"required\": [],\n" +
            "                        \"additionalProperties\": {\n" +
            "                            \"anyOf\": [\n" +
            "                                {\n" +
            "                                    \"type\": \"string\"\n" +
            "                                },\n" +
            "                                {\n" +
            "                                    \"type\": \"integer\"\n" +
            "                                }\n" +
            "                            ]\n" +
            "                        }\n" +
            "                    }\n" +
            "                },\n" +
            "                {\n" +
            "                    \"type\": \"string\",\n" +
            "                    \"pattern\": \"" + strategyJsonPattern + "\"\n" +
            "                }\n" +
            "            ]\n" +
            "        },\n" +
            "        \"exclude\": {\n" +
            "            \"description\": \"值格式为：List,用于排除 matrix  中的一些组合,每个元素为一个 Object,或是一个'$'{{fromJSON(xxx)}}上下文\",\n" +
            "            \"oneOf\": [\n" +
            "                {\n" +
            "                    \"type\": \"array\",\n" +
            "                    \"items\": {\n" +
            "                        \"type\": \"object\",\n" +
            "                        \"required\": [],\n" +
            "                        \"additionalProperties\": {\n" +
            "                            \"anyOf\": [\n" +
            "                                {\n" +
            "                                    \"type\": \"string\"\n" +
            "                                },\n" +
            "                                {\n" +
            "                                    \"type\": \"integer\"\n" +
            "                                }\n" +
            "                            ]\n" +
            "                        }\n" +
            "                    }\n" +
            "                },\n" +
            "                {\n" +
            "                    \"type\": \"string\",\n" +
            "                    \"pattern\": \"" + strategyJsonPattern + "\"\n" +
            "                }\n" +
            "            ]\n" +
            "        },\n" +
            "        \"strategy\": {\n" +
            "            \"description\": \"值格式为：Object,,或是一个'$'{{fromJSON(xxx)}}上下文,定义的每个选项都有键和值，键将作为 matrix 上下文中的属性\",\n" +
            "            \"oneOf\": [\n" +
            "                {\n" +
            "                    \"type\": \"object\",\n" +
            "                    \"additionalProperties\": {\n" +
            "                        \"anyOf\": [\n" +
            "                            {\n" +
            "                                \"type\": \"array\"\n" +
            "                            },\n" +
            "                            {\n" +
            "                                \"type\": \"string\",\n" +
            "                                \"pattern\": \"" + strategyJsonPattern + "\"\n" +
            "                            }\n" +
            "                        ]\n" +
            "                    }\n" +
            "                },\n" +
            "                {\n" +
            "                    \"type\": \"string\",\n" +
            "                    \"pattern\": \"" + strategyJsonPattern + "\"\n" +
            "                }\n" +
            "            ]\n" +
            "        }\n" +
            "    }\n" +
            "}";

    private static final ThreadLocal<Yaml> yaml = ThreadLocal.withInitial(Yaml::new);

    private static final JsonSchemaFactory schemaFactory = JsonSchemaFactory
            .builder(JsonSchemaFactory.getInstance(SpecVersion.VersionFlag.V7))
            .yamlMapper(YamlUtil.objectMapper)
            .build();

    /**
     * 使用循环遍历笛卡尔乘积算法
     */
    public static List<List<Object>> loopCartesianProduct(List<List<String>> input) {
        List<List<Object>> acc = new ArrayList<>();
        acc.add(new ArrayList<>());
        for (List<String> nextList : input) {
            List<List<Object>> newAcc = new ArrayList<>();
            for (List<Object> list : acc) {
                for (Object element : nextList) {
                    List<Object> newList = new ArrayList<>(list);
                    newList.add(element);
                    newAcc.add(newList);
                }
            }
            acc = newAcc;
        }
        return acc;
    }

    /**
     * 使用递归遍历实现的笛卡尔乘积算法，因为容易导致堆栈溢出，仅用于单测校验
     */
    public static List<List<Object>> recursiveCartesianProduct(List<List<Object>> input) {
        List<List<Object>> output = new ArrayList<>();
        product(input, output, 0, new ArrayList<>());
        return output;
    }

    public static void schemaCheck(String originYaml) {
        if (originYaml.isBlank()) {
            return;
        }
        JsonNode yamlJson = YamlUtil.objectMapper.convertValue(yaml.get().load(originYaml), JsonNode.class);
        yamlJson = replaceOn(yamlJson);
        JsonSchema schema = schemaFactory.getSchema(schemaJson);
        check(schema, yamlJson);
    }

    private static void check(JsonSchema schema, JsonNode yaml) {
        Set<ValidationMessage> errors = schema.validate(yaml);
        if (!errors.isEmpty()) {
            throw new ServerException(-1000, errors.toString());
        }
    }

    // Yaml规则下会将on当成true在消除锚点时会将On替换为true
    private static JsonNode replaceOn(JsonNode node) {
        JsonNode realOn = node.get("true");
        if (realOn == null) {
            return node;
        }
        ObjectNode objectNode = (ObjectNode) node;
        objectNode.set("on", realOn);
        objectNode.remove("true");
        return objectNode;
    }

    /**
     * 笛卡尔乘积递归遍历操作:
     * 原二维数组[input], 通过乘积转化后的数组[output],
     * 层级参数[layer], 当前操作数组[currentList]
     */
    private static void product(List<List<Object>> input, List<List<Object>> output, int layer, List<Object> currentList) {
        if (layer < input.size() - 1) {
            if (input.get(layer).isEmpty()) {
                product(input, output, layer + 1, currentList);
            } else {
                for (Object element : input.get(layer)) {
                    List<Object> newList = new ArrayList<>(currentList);
                    newList.add(element);
                    product(input, output, layer + 1, newList);
                }
            }
        } else if (layer == input.size() - 1) {
            if (input.get(layer).isEmpty()) {
                output.add(currentList);
            } else {
                for (Object element : input.get(layer)) {
                    List<Object> newList = new ArrayList<>(currentList);
                    newList.add(element);
                    output.add(newList);
                }
            }
        }
    }
}
