package cd.casic.ci.log.service;

import cd.casic.ci.log.dal.dao.LogTahMapper;
import cd.casic.framework.commons.util.json.JsonUtils;
import cd.casic.framework.redis.core.RedisTemplateUtils;
import com.fasterxml.jackson.core.type.TypeReference;
import jakarta.annotation.Resource;
import lombok.extern.slf4j.Slf4j;
import org.apache.logging.log4j.util.Strings;
import org.redisson.RedissonRedLock;
import org.redisson.api.RLock;
import org.redisson.api.RedissonClient;
import org.springframework.stereotype.Service;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;

/**
 * @author by mianbin
 * @Classname LogTagService
 * @Description TODO
 * @Date 2025/3/20 14:49
 */
@Slf4j
@Service
public class LogTagService {

    @Resource
    private LogTahMapper logTahMapper;
    @Resource
    private RedisTemplateUtils redisTemplateUtils;
    @Resource
    private RedissonClient redissonClient;

    private static String LOG_SUBTAG = "log:build:tag:subTags:";
    private static String LOG_SUBTAG_LOCK = "log:build:tag:subTags:distribute:lock:";

    public static String getSubTagsRedisKey(String buildId, String tagName) {
        return LOG_SUBTAG + genBuildIdAndTagKey(buildId, tagName);
    }

    private static String genBuildIdAndTagKey(String buildId, String tagName) {
        return buildId + ":" + tagName;
    }

    public void saveSubTag(String buildId, String tagName, String subTag) {
        // 从 redis 和数据库里面 看是否已经保存
        List<String> originSubTags = getSubTags(buildId, tagName);
        // 如果已经保存了
        if (originSubTags != null && originSubTags.contains(subTag)) {
            return;
        }

        // 如果没有保存，也就是数据库里面没有
        List<String> subTags = new ArrayList<>();
        subTags.add(subTag);

        if (originSubTags != null) {
            subTags.addAll(originSubTags);
        }

        // 存入数据库
        logTahMapper.save(buildId, tagName, JsonUtils.toJsonString(subTags));
        saveSubTagsToRedis(buildId, tagName, JsonUtils.toJsonString(subTags));

        log.info("[" + buildId + "|" + tagName + "] Create new subTag in db and redis: " + subTag);
    }

    private List<String> getSubTags(String buildId, String tagName) {
        // 看 redis 里面有没有
        String subTagsStr = (String) redisTemplateUtils.get(getSubTagsRedisKey(buildId, tagName));

        if (Strings.isBlank(subTagsStr)) {
            // redis 里面没有就看数据库里面有木有
            String subTags = logTahMapper.getSubTags(buildId, tagName);
            if (subTags != null) {
                // 如果有，就存入 redis 并且返回
                saveSubTagsToRedis(buildId, tagName, subTags);
                return JsonUtils.parseObject(subTags, new TypeReference<List<String>>() {
                });
            }
            // 没有返回 null
            return null;
        } else {
            return JsonUtils.parseObject(subTagsStr, new TypeReference<List<String>>() {
            });
        }
    }

    private void saveSubTagsToRedis(String buildId, String tagName, String subTags) {
        RLock lock = redissonClient.getLock(LOG_SUBTAG_LOCK);
        RedissonRedLock redisLock = new RedissonRedLock(lock);
        try {
            redisLock.lock();
            redisTemplateUtils.set(getSubTagsRedisKey(buildId, tagName), subTags, TimeUnit.DAYS.toSeconds(1));
        } finally {
            redisLock.unlock();
        }
    }

}
