package cd.casic.ci.process.api.engine.common;

import cd.casic.ci.common.pipeline.enums.BuildRecordTimeStamp;
import cd.casic.ci.common.pipeline.pojo.time.BuildRecordTimeCost;
import cd.casic.ci.common.pipeline.pojo.time.BuildRecordTimeLine;
import cd.casic.ci.process.api.process.pojo.pipeline.record.BuildRecordContainer;
import cd.casic.ci.process.api.process.pojo.pipeline.record.BuildRecordModel;
import cd.casic.ci.process.api.process.pojo.pipeline.record.BuildRecordStage;
import cd.casic.ci.process.api.process.pojo.pipeline.record.BuildRecordTask;
import cn.hutool.core.lang.Pair;
import cn.hutool.json.JSONUtil;
import lombok.extern.slf4j.Slf4j;

import java.time.Duration;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.List;

/**
 * @author by mianbin
 * @Classname BuildTimeCostUtils
 * @Description TODO
 * @Date 2025/3/31 16:07
 */
@Slf4j
public class BuildTimeCostUtils {

    public static BuildRecordTimeCost generateBuildTimeCost(BuildRecordModel buildRecordModel, List<BuildRecordStage> stageRecords) {
        LocalDateTime startTime = buildRecordModel.getStartTime();
        if (startTime == null) {
            return new BuildRecordTimeCost();
        }
        LocalDateTime endTime = buildRecordModel.getEndTime() != null ? buildRecordModel.getEndTime() : LocalDateTime.now();
        long totalCost = Duration.between(buildRecordModel.getQueueTime(), endTime).toMillis();
        long executeCost = 0L;
        long waitCost = 0L;
        long queueCost = 0L;
        for (BuildRecordStage record : stageRecords) {
            //这里转换需要注意，要根据实际发来的进行获取, 这里获取的应该是stage里面的BuildRecordTimeCost，要调试
            String stageVar = (String) record.getStageVar().get(BuildRecordTimeCost.class.getSimpleName());
            if (stageVar == null) {
                continue;
            }
            BuildRecordTimeCost stageCost = JSONUtil.toBean(stageVar, BuildRecordTimeCost.class, false);
            executeCost += stageCost.getExecuteCost();
            waitCost += stageCost.getWaitCost();
            queueCost += stageCost.getQueueCost();
            if (stageCost == null) {
                continue;
            }
            executeCost += stageCost.getExecuteCost();
            waitCost += stageCost.getWaitCost();
            queueCost += stageCost.getQueueCost();
        }
        long systemCost = totalCost - executeCost - queueCost - waitCost;
        return new BuildRecordTimeCost(
                totalCost,
                executeCost,
                waitCost,
                Duration.between(buildRecordModel.getQueueTime(), startTime).toMillis(),
                notNegative(systemCost)
        );
    }

    public static BuildRecordTimeCost generateStageTimeCost(BuildRecordStage buildRecordStage, List<BuildRecordContainer> containerRecords) {
        LocalDateTime startTime = buildRecordStage.getStartTime();
        if (startTime == null) {
            return null;
        }
        LocalDateTime endTime = buildRecordStage.getEndTime() != null ? buildRecordStage.getEndTime() : LocalDateTime.now();
        long totalCost = Duration.between(startTime, endTime).toMillis();
        List<BuildRecordTimeLine.Moment> containerExecuteCost = new ArrayList<>();
        List<BuildRecordTimeLine.Moment> containerWaitCost = List.of(
                new BuildRecordTimeLine.Moment(startTime, endTime)
        );
        List<BuildRecordTimeLine.Moment> containerQueueCost = List.of(
                new BuildRecordTimeLine.Moment(startTime, endTime)
        );
        for (BuildRecordContainer record : containerRecords) {
            BuildRecordTimeLine containerTimeLine = JSONUtil.toBean( record.getContainerVar().get(BuildRecordTimeLine.class.getSimpleName()).toString(), BuildRecordTimeLine.class, false);
            if (containerTimeLine == null) {
                continue;
            }
            if (record.getEndTime() != null) {
                BuildRecordTimeLine.Moment fixedMoment = new BuildRecordTimeLine.Moment(record.getEndTime(), endTime);
                containerTimeLine.getWaitCostMoments().add(fixedMoment);
                containerTimeLine.getQueueCostMoments().add(fixedMoment);
            }
            containerExecuteCost = mergeTimeLine(containerExecuteCost, containerTimeLine.getExecuteCostMoments());
            List<BuildRecordTimeLine.Moment> mergedWaitCost = mergeTimeLine(containerTimeLine.getWaitCostMoments(), containerTimeLine.getQueueCostMoments());
            containerWaitCost = intersectionTimeLine(containerWaitCost, mergedWaitCost);
            containerQueueCost = intersectionTimeLine(containerQueueCost, containerTimeLine.getQueueCostMoments());
        }
        long executeCost = containerExecuteCost.stream().mapToLong(moment -> moment.getEndTime() - moment.getStartTime()).sum();
        long queueCost = containerQueueCost.stream().mapToLong(moment -> moment.getEndTime() - moment.getStartTime()).sum();
        long waitCost = buildRecordStage.getTimestamps().entrySet().stream()
                .filter(entry -> entry.getKey().stageCheckWait())
                .mapToLong(entry -> {
                    BuildRecordTimeStamp time = entry.getValue();
                    logWhenNull(time, buildRecordStage.getBuildId() + "|STAGE|" + buildRecordStage.getStageId() + "|" + entry.getKey().name());
                    return time.between();
                })
                .sum() + containerWaitCost.stream().mapToLong(moment -> moment.getEndTime() - moment.getStartTime()).sum();
        long systemCost = totalCost - executeCost - waitCost;
        return new BuildRecordTimeCost(
                totalCost,
                executeCost,
                waitCost,
                queueCost,
                notNegative(systemCost)
        );
    }

    public static BuildRecordTimeCost generateMatrixTimeCost(BuildRecordContainer buildRecordContainer, List<BuildRecordContainer> containerRecords) {
        LocalDateTime startTime = buildRecordContainer.getStartTime();
        if (startTime == null) {
            return null;
        }
        LocalDateTime endTime = buildRecordContainer.getEndTime() != null ? buildRecordContainer.getEndTime() : LocalDateTime.now();
        long totalCost = Duration.between(startTime, endTime).toMillis();
        List<BuildRecordTimeLine.Moment> containerExecuteCost = new ArrayList<>();
        List<BuildRecordTimeLine.Moment> containerWaitCost = List.of(
                new BuildRecordTimeLine.Moment(startTime, endTime)
        );
        List<BuildRecordTimeLine.Moment> containerQueueCost = List.of(
                new BuildRecordTimeLine.Moment(startTime, endTime)
        );
        for (BuildRecordContainer record : containerRecords) {
            BuildRecordTimeLine containerTimeLine = JSONUtil.toBean( record.getContainerVar().get(BuildRecordTimeLine.class.getSimpleName()).toString(), BuildRecordTimeLine.class, false);
            if (containerTimeLine == null) {
                continue;
            }
            containerExecuteCost = mergeTimeLine(containerExecuteCost, containerTimeLine.getExecuteCostMoments());
            containerWaitCost = intersectionTimeLine(containerWaitCost, containerTimeLine.getWaitCostMoments());
            containerQueueCost = intersectionTimeLine(containerQueueCost, containerTimeLine.getQueueCostMoments());
        }
        long executeCost = containerExecuteCost.stream().mapToLong(moment -> moment.getEndTime() - moment.getStartTime()).sum();
        long queueCost = containerQueueCost.stream().mapToLong(moment -> moment.getEndTime() - moment.getStartTime()).sum();
        long waitCost = containerWaitCost.stream().mapToLong(moment -> moment.getEndTime() - moment.getStartTime()).sum();
        long systemCost = totalCost - executeCost - queueCost - waitCost;
        return new BuildRecordTimeCost(
                totalCost,
                executeCost,
                waitCost,
                queueCost,
                notNegative(systemCost)
        );
    }

    public static Pair<BuildRecordTimeCost, BuildRecordTimeLine> generateContainerTimeCost(BuildRecordContainer buildRecordContainer, List<BuildRecordTask> taskRecords) {
        BuildRecordTimeLine containerTimeLine = new BuildRecordTimeLine();
        LocalDateTime startTime = buildRecordContainer.getStartTime();
        if (startTime == null) {
            startTime = taskRecords.isEmpty() ? null : taskRecords.get(0).getStartTime();
            if (startTime == null) {
                return Pair.of(null, containerTimeLine);
            }
        }
        LocalDateTime endTime = buildRecordContainer.getEndTime() != null ? buildRecordContainer.getEndTime() : LocalDateTime.now();
        long totalCost = Duration.between(startTime, endTime).toMillis();
        long executeCost = 0L;
        long waitCost = 0L;
        long queueCost = buildRecordContainer.getTimestamps().entrySet().stream()
                .filter(entry -> entry.getKey().containerCheckQueue())
                .mapToLong(entry -> {
                    BuildRecordTimeStamp time = entry.getValue();
                    logWhenNull(time, buildRecordContainer.getBuildId() + "|CONTAINER|" + buildRecordContainer.getContainerId() + "|" + entry.getKey().name());
                    time.insert2TimeLine(containerTimeLine.getQueueCostMoments());
                    return time.between();
                })
                .sum();
        for (BuildRecordTask record : taskRecords) {
            BuildRecordTimeLine taskTimeLine = new BuildRecordTimeLine();
            BuildRecordTimeCost cost = generateTaskTimeCost(record, taskTimeLine);
            if (cost != null) {
                containerTimeLine.getQueueCostMoments().addAll(taskTimeLine.getQueueCostMoments());
                containerTimeLine.getWaitCostMoments().addAll(taskTimeLine.getWaitCostMoments());
                containerTimeLine.getExecuteCostMoments().addAll(taskTimeLine.getExecuteCostMoments());
                executeCost += cost.getExecuteCost();
                waitCost += cost.getWaitCost();
            }
        }
        long systemCost = totalCost - executeCost - waitCost - queueCost;
        return Pair.of(
                new BuildRecordTimeCost(
                        totalCost,
                        executeCost,
                        waitCost,
                        queueCost,
                        notNegative(systemCost)
                ),
                containerTimeLine
        );
    }

    public static BuildRecordTimeCost generateTaskTimeCost(BuildRecordTask buildRecordTask, BuildRecordTimeLine timeLine) {
        LocalDateTime startTime = buildRecordTask.getStartTime();
        if (startTime == null) {
            return null;
        }
        LocalDateTime endTime = buildRecordTask.getEndTime() != null ? buildRecordTask.getEndTime() : LocalDateTime.now();
        long totalCost = Duration.between(startTime, endTime).toMillis();
        long waitCost = buildRecordTask.getTimestamps().entrySet().stream()
                .filter(entry -> entry.getKey().taskCheckWait())
                .mapToLong(entry -> {
                    BuildRecordTimeStamp time = entry.getValue();
                    logWhenNull(time, buildRecordTask.getBuildId() + "|TASK|" + buildRecordTask.getTaskId() + "|" + entry.getKey().name());
                    if (timeLine != null) {
                        time.insert2TimeLine(timeLine.getWaitCostMoments());
                    }
                    return time.between();
                })
                .sum();
        long executeCost = totalCost - waitCost;
        if (timeLine != null) {
            List<BuildRecordTimeLine.Moment> executeMoments = differenceTimeLine(
                    List.of(new BuildRecordTimeLine.Moment(startTime, endTime)),
                    timeLine.getWaitCostMoments()
            );
            timeLine.getExecuteCostMoments().addAll(executeMoments);
        }
        return new BuildRecordTimeCost(
                totalCost,
                waitCost,
                notNegative(executeCost)
        );
    }

    public static List<BuildRecordTimeLine.Moment> differenceTimeLine(List<BuildRecordTimeLine.Moment> left, List<BuildRecordTimeLine.Moment> right) {
        List<Pair<Long, Character>> line = new ArrayList<>();
        List<BuildRecordTimeLine.Moment> ans = new ArrayList<>();
        for (BuildRecordTimeLine.Moment moment : left) {

            line.add(Pair.of(moment.getStartTime(), 'L'));
            line.add(Pair.of(moment.getEndTime(), 'L'));
        }
        for (BuildRecordTimeLine.Moment moment : right) {
            line.add(Pair.of(moment.getStartTime(), 'R'));
            line.add(Pair.of(moment.getEndTime(), 'R'));
        }
        line.sort(java.util.Comparator.comparingLong(Pair::getKey));
        boolean cnt = true;
        int index = 0;
        while (index < line.size()) {
            if (line.get(index).getValue() == 'R') {
                cnt = !cnt;
            }
            if (cnt && index < line.size() - 1) {
                ans.add(new BuildRecordTimeLine.Moment(line.get(index).getKey(), line.get(index + 1).getKey()));
            }
            index++;
        }
        return ans;
    }

    public static List<BuildRecordTimeLine.Moment> intersectionTimeLine(List<BuildRecordTimeLine.Moment> left, List<BuildRecordTimeLine.Moment> right) {
        List<BuildRecordTimeLine.Moment> ans = new ArrayList<>();
        int i = 0;
        int j = 0;
        while (i < left.size() && j < right.size()) {
            long lo = Math.max(left.get(i).getStartTime(), right.get(j).getStartTime());
            long hi = Math.min(left.get(i).getEndTime(), right.get(j).getEndTime());
            if (lo <= hi) {
                ans.add(new BuildRecordTimeLine.Moment(lo, hi));
            }
            if (left.get(i).getEndTime() < right.get(j).getEndTime()) {
                i++;
            } else {
                j++;
            }
        }
        return ans;
    }

    public static List<BuildRecordTimeLine.Moment> mergeTimeLine(List<BuildRecordTimeLine.Moment> left, List<BuildRecordTimeLine.Moment> right) {
        List<BuildRecordTimeLine.Moment> intervals = new ArrayList<>(left);
        intervals.addAll(right);
        intervals.sort(java.util.Comparator.comparingLong(BuildRecordTimeLine.Moment::getStartTime));
        List<BuildRecordTimeLine.Moment> res = new ArrayList<>();
        for (BuildRecordTimeLine.Moment interval : intervals) {
            if (res.isEmpty() || res.get(res.size() - 1).getEndTime() < interval.getStartTime()) {
                res.add(interval);
            } else {
                BuildRecordTimeLine.Moment m = res.remove(res.size() - 1);
                res.add(new BuildRecordTimeLine.Moment(m.getStartTime(), Math.max(m.getEndTime(), interval.getEndTime())));
            }
        }
        return res;
    }

    private static void logWhenNull(BuildRecordTimeStamp time, String logInfo) {
        if (time.getStartTime() == null) {
            log.warn(logInfo + "|warning! start time is null.");
        }
        if (time.getEndTime() == null) {
            log.warn(logInfo + "|warning! end time is null.");
        }
    }

    private static long notNegative(long value) {
        return Math.max(value, 0);
    }
}
