package cd.casic.ci.log.scm.service.tgit;

import cd.casic.framework.commons.util.encrypt.AESUtil;
import cd.casic.ci.log.scm.dal.dao.RepositoryTGitTokenMapper;
import cd.casic.ci.log.scm.dal.dto.RepositoryTGitToken;
import cd.casic.ci.log.scm.dal.pojo.oauth.GitToken;
import cd.casic.ci.log.scm.enums.TokenTypeEnum;
import jakarta.annotation.Resource;
import lombok.extern.slf4j.Slf4j;
import org.redisson.RedissonRedLock;
import org.redisson.api.RLock;
import org.redisson.api.RedissonClient;
import org.springframework.stereotype.Service;

import java.time.ZoneId;


@Service
@Slf4j
public class TGitTokenServiceImpl implements TGitTokenService {

    private final String aesKey = "";

    @Resource
    private RedissonClient redissonClient;

    @Resource
    private ITGitService tGitService;

    @Resource
    private RepositoryTGitTokenMapper repositoryTGitTokenMapper;

    @Override
    public GitToken getAccessToken(String userId) {
        GitToken accessToken = doGetAccessToken(userId);
        if (accessToken == null) {
            return null;
        }
        if (isTokenExpire(accessToken)) {
            log.info("try to refresh the git token of user({})", userId);
            RLock rLock = redissonClient.getLock("OAUTH_REFRESH_TOKEN_" + userId);
            RedissonRedLock lock = new RedissonRedLock(rLock);

            try {
                lock.lock();
                GitToken newAccessToken = doGetAccessToken(userId);
                assert newAccessToken != null;
                if (newAccessToken.getExpiresIn() * 1000 <= System.currentTimeMillis() - 1800 * 1000) {
                    return refreshToken(userId, newAccessToken);
                } else {
                    return newAccessToken;
                }
            } finally {
                lock.unlock();
            }
        } else {
            return accessToken;
        }
    }

    @Override
    public Integer saveAccessToken(String userId, String oauthUserId, GitToken tGitToken) {
        tGitToken.setAccessToken(AESUtil.encrypt(aesKey, tGitToken.getAccessToken()));
        tGitToken.setRefreshToken(AESUtil.encrypt(aesKey, tGitToken.getRefreshToken()));
        RepositoryTGitToken repositoryTGitToken = new RepositoryTGitToken();
        repositoryTGitToken.setUserId(userId);
        repositoryTGitToken.setOauthUserId(oauthUserId);
        repositoryTGitToken.setAccessToken(tGitToken.getAccessToken());
        repositoryTGitToken.setRefreshToken(tGitToken.getRefreshToken());
        repositoryTGitToken.setTokenType(tGitToken.getTokenType());
        repositoryTGitToken.setExpiresIn(tGitToken.getExpiresIn());
        return repositoryTGitTokenMapper.insert(repositoryTGitToken);
    }

    private GitToken doGetAccessToken(String userId) {
        RepositoryTGitToken token = repositoryTGitTokenMapper.selectOne(RepositoryTGitToken::getUserId, userId);
        if (token == null) {
            return null;
        }
        return new GitToken(AESUtil.decrypt(aesKey, token.getAccessToken()),
                AESUtil.decrypt(aesKey, token.getRefreshToken()),
                token.getTokenType(), token.getExpiresIn(), token.getCreateTime().atZone(ZoneId.systemDefault()).toInstant().toEpochMilli());
    }

    private GitToken refreshToken(String userId, GitToken gitToken) {
        GitToken token = tGitService.refreshToken(userId, gitToken);
        String username = tGitService.getUserInfoByToken(token.getAccessToken(), TokenTypeEnum.OAUTH).getUsername();
        String oauthUserId = username != null ? username : userId;
        saveAccessToken(userId, oauthUserId, token);
        token.setAccessToken(AESUtil.decrypt(aesKey, token.getAccessToken()));
        token.setRefreshToken(AESUtil.decrypt(aesKey, token.getRefreshToken()));
        return token;
    }

    private Boolean isTokenExpire(GitToken accessToken) {
        // 提前半个小时刷新token
        return (accessToken.getCreateTime() != null ? accessToken.getCreateTime() : 0) + accessToken.getExpiresIn() * 1000 - 1800 * 1000 <= System.currentTimeMillis();
    }
}
